import json
from os import environ
from pathlib import Path
from typing import List, Union

import hvac
from hvac.api.auth_methods import Kubernetes
from hvac.exceptions import Forbidden, InvalidPath, InvalidRequest
from loguru import logger
from pydantic import ValidationError

from vault_sync.config import AuthMethod, Config, Vault
from vault_sync.exceptions import ConfigException, VaultLoginException


class VaultSync:
    """
    A class to synchronize secrets between a source Vault and a destination Vault.

    This class is responsible for automating the process of retrieving, parsing, and
    transferring structured data (secrets) from a source HashiCorp Vault instance
    to a destination HashiCorp Vault instance. It supports listing secrets,
    retrieving paths, fetching secrets, and updating destination secrets with
    the intention of maintaining consistent state between the two Vault instances.
    """

    def __init__(self, config_file: str):
        self.config = self.load_config(config_file)
        self.source_client = hvac.Client(url=self.config.source.url)
        self.destination_client = hvac.Client(url=self.config.destination.url)

    @staticmethod
    def load_config(config_file: Union[Path, str]) -> Config:
        """
        Loads a configuration file, parses its contents, and returns a Config object.

        This method expects a JSON format for the configuration data and attempts to
        deserialize it into a Config object. It validates the configuration format
        using strict adherence to rules defined in the Config data model.

        Arguments:
        config_file: Path or str
            Path to the configuration file on disk, given as either a pathlib.Path
            object or a string representing file path.

        Returns:
        Config
            A fully initialized and validated Config object created using the
            parsed JSON data.

        Raises:
        ConfigException
            If the content of the provided file is not valid JSON or fails
            to match the validation criteria imposed by the Config model.
        """
        try:
            with open(config_file) as json_file:
                data = json.load(json_file)
            config = Config(**data)
            return config
        except (json.JSONDecodeError, ValidationError) as err:
            raise ConfigException(str(err))

    @staticmethod
    def authenticate_client(client: hvac.Client, config: Vault) -> None:
        if config.auth_method == AuthMethod.APPROLE:
            client.auth.approle.login(
                role_id=config.role_id,
                secret_id=config.secret_id,
            )
        elif config.auth_method == AuthMethod.KUBERNETES:
            try:
                with open(config.token_path) as token_file:
                    token = token_file.read()
            except FileNotFoundError:
                raise VaultLoginException("token file not found.")

            Kubernetes(client.adapter).login(config.role_name, token)
        else:
            client.token = config.token
            if environ.get("VAULT_TOKEN"):
                client.token = environ.get("VAULT_TOKEN")

        # raise if we are still not authenticated
        if not client.is_authenticated():
            raise VaultLoginException("failed to authenticate.")

    def authenticate_clients(self) -> None:
        """
        Authenticate clients with Vault services.

        This method attempts to authenticate both source and destination clients
        using approle-based login. If the clients are already authenticated, the
        method exits without re-authentication.

        Raises:
            VaultLoginException: Raised when authentication fails due to invalid
                requests or forbidden access.

        """
        try:
            if self.source_client.is_authenticated() and self.destination_client.is_authenticated():
                logger.debug("clients already authenticated.")
                return

            logger.info("signin to source vault...")
            self.authenticate_client(client=self.source_client, config=self.config.source)

            logger.info("signin to destination vault...")
            self.authenticate_client(client=self.destination_client, config=self.config.source)
        except (InvalidRequest, Forbidden) as err:
            raise VaultLoginException(str(err))

    def list_keys(self, kv_store_client: hvac.Client, kv_path: str) -> List[str]:
        """
        Retrieve and filter the list of keys from a specified key-value (KV) store path.

        This function interacts with a KV secrets engine in a configured Vault instance
        to retrieve the list of keys at a given path. The keys are filtered to exclude
        those that end with a forward slash ("/"), which typically represent directories
        in the KV store.

        Parameters
        ----------
        kv_store_client : hvac.Client
            The client instance used to connect to the Vault server, configured for the
            appropriate authentication method and permissions.
        kv_path : str
            The path inside the key-value store from which the keys are to be listed.

        Returns
        -------
        List[str]
            A list of keys present at the specified KV path, excluding those that end
            with a forward slash.
        """
        list_response = kv_store_client.secrets.kv.v2.list_secrets(
            path=kv_path, mount_point=self.config.source.kv_store
        )
        return [key for key in list_response["data"]["keys"] if not key.endswith("/")]

    def list_folders(self, kv_store_client: hvac.Client, kv_path: str) -> List[str]:
        """
        Retrieves a list of folder names from a KV store using the given KV store client.

        This function interacts with a HashiCorp Vault KV v2 secrets engine to list the
        folders (keys ending with '/') under a specific path. It filters and returns only
        the folder names from the retrieved data.

        Args:
            kv_store_client (hvac.Client): The configured HVAC client for interacting with
                                           the KV store.
            kv_path (str): The path within the KV store where the secrets are located.

        Returns:
            List[str]: A list of folder names (keys ending with '/') under the specified
                       KV path.
        """
        list_response = kv_store_client.secrets.kv.v2.list_secrets(
            path=kv_path, mount_point=self.config.source.kv_store
        )
        return [key for key in list_response["data"]["keys"] if key.endswith("/")]

    def list_all_paths_in_source(self) -> List[str]:
        """
        Retrieves all possible file or directory paths from the source client.

        This method uses the `walk_paths` functionality to traverse the source client
        configuration and returns a list containing all identified paths.

        Returns
        -------
        List[str]
            A list of string representations of all paths found in the source client.
        """
        return self.walk_paths(self.source_client, "", [""])

    def walk_paths(self, client: hvac.Client, root: str, folders: List[str]) -> List[str]:
        """
        Recursively walks through a hierarchy of paths in a Vault-like storage and
        collects all paths into a list. This method starts from a root path and
        explores all nested folders accumulating their paths.

        Parameters:
            client (hvac.Client): The client instance used to interact with the storage system.
            root (str): The root path from which to start walking through the hierarchy.
            folders (List[str]): A list to store the paths of all discovered folders.

        Returns:
            List[str]: A list containing the paths of all folders reachable from the root path.
        """
        paths = [f"{root}{folder}" for folder in self.list_folders(client, root)]
        folders.extend(paths)
        for folder in paths:
            self.walk_paths(client, folder, folders)
        return folders

    def list_all_keys_in_source(self) -> List[str]:
        """
        Summarize and list all keys from the specified source paths.

        The method retrieves all base paths defined in the source, iterates through them,
        and gathers every key from each specified path. The resulting list provides an
        aggregated view of all keys that can be accessed from the given source structure.

        Returns:
            List[str]: A list containing the aggregated keys as strings.
        """
        return [
            f"{explore}{key}"
            for explore in self.list_all_paths_in_source()
            for key in self.list_keys(self.source_client, explore)
        ]

    def get_secret_from_source(self, secret_path: str) -> dict | None:
        """
        Retrieves a secret from a specified source path.

        This method fetches the secret stored at the provided path using the
        configured source client. The retrieved secret is expected to be returned
        in the form of a dictionary containing the data.

        Parameters:
            secret_path (str): The path to the secret in the source storage.

        Returns:
            dict: A dictionary containing the secret data.
        """
        try:
            secret_version_response = self.source_client.secrets.kv.v2.read_secret_version(
                path=secret_path, mount_point=self.config.source.kv_store
            )
        except InvalidPath:
            logger.debug(f"secret {secret_path} not found in source, skipping...")
            return None

        return secret_version_response["data"]["data"]

    def update_destination_secret(self, secret_path: str, secret_data: dict) -> None:
        """
        Updates or creates a secret at the destination path.

        This function checks if a secret exists at the specified destination path and
        compares it with the provided secret data. If the secret exists and the data
        matches, the update is skipped. If the secret does not exist or the data
        differs, the secret is updated or created accordingly.

        Args:
            secret_path: The path to the secret in the destination key-value store.
            secret_data: The data to be stored in the secret.

        Raises:
            InvalidPath: If the specified secret path is invalid.

        Returns:
            None
        """
        try:
            # if secret is not None:
            secret_version_response = self.destination_client.secrets.kv.v2.read_secret_version(
                path=secret_path, mount_point=self.config.destination.kv_store
            )

            if secret_version_response["data"]["data"] == secret_data:
                logger.debug(f"skipping {secret_path}, secret is up to date.")
                return

            logger.info(f"secret {secret_path} in destination differs, updating...")
            exists = True
        except InvalidPath:
            exists = False
            logger.info(f"secret {secret_path} not found in destination, creating...")

        if exists and secret_data is None:
            logger.info(f"secret {secret_path} not found in source but found in destination, deleting...")
            self.destination_client.secrets.kv.v2.delete_metadata_and_all_versions(
                path=secret_path, mount_point=self.config.destination.kv_store, secret=secret_data
            )
        else:
            self.destination_client.secrets.kv.v2.create_or_update_secret(
                path=secret_path, mount_point=self.config.destination.kv_store, secret=secret_data
            )

    def sync(self, secret_path: str) -> None:
        """
        Synchronizes the secret at the specified path with the source.

        This method retrieves a secret from a source location and updates its
        destination counterpart with the fetched data. It is designed to maintain
        synchronization between the source secret and its corresponding destination
        secret.

        Parameters:
            secret_path (str): The path of the secret to be synchronized.

        Returns:
            None
        """
        self.update_destination_secret(secret_path, self.get_secret_from_source(secret_path))

    def sync_all(self) -> None:
        """
        Synchronizes all secrets between source and destination.

        This method iterates over all keys retrieved from the source and
        synchronizes each one using the `sync` method. Upon completion,
        a log message is generated to indicate the process has finished.

        Returns:
            None
        """
        for secret_path in self.list_all_keys_in_source():
            self.sync(secret_path)
        logger.info("all done.")
